@@ -12,20 +12,9 @@
 #include <ext/apcu/apc_serializer.h>
 #include <zend_smart_str.h>
 #endif
+#include <Zend/zend_interfaces.h>
 #include "php_brotli.h"
 
-#ifndef TSRMLS_C
-#define TSRMLS_C
-#endif
-#ifndef TSRMLS_CC
-#define TSRMLS_CC
-#endif
-#ifndef TSRMLS_DC
-#define TSRMLS_DC
-#endif
-
-int le_state;
-
 # pragma GCC diagnostic ignored "-Wpointer-sign"
 
 ZEND_DECLARE_MODULE_GLOBALS(brotli);
@@ -106,7 +95,7 @@ static int php_brotli_encoder_create(BrotliEncoderState **encoder,
     }
 
     if (quality < BROTLI_MIN_QUALITY || quality > BROTLI_MAX_QUALITY) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
+        php_error_docref(NULL, E_WARNING,
                          "brotli: compression level (%ld) "
                          "must be within %d..%d",
                          (long)quality, BROTLI_MIN_QUALITY, BROTLI_MAX_QUALITY);
@@ -118,7 +107,7 @@ static int php_brotli_encoder_create(BrotliEncoderState **encoder,
     if (mode != BROTLI_MODE_GENERIC &&
         mode != BROTLI_MODE_TEXT &&
         mode != BROTLI_MODE_FONT) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
+        php_error_docref(NULL, E_WARNING,
                          "brotli: compression mode (%ld) must be %d, %d, %d",
                          (long)mode, BROTLI_MODE_GENERIC, BROTLI_MODE_TEXT,
                          BROTLI_MODE_FONT);
@@ -142,35 +131,6 @@ static int php_brotli_decoder_create(BrotliDecoderState **decoder)
     return SUCCESS;
 }
 
-static php_brotli_state_context* php_brotli_state_init(void)
-{
-    php_brotli_state_context *ctx
-        = (php_brotli_state_context *)ecalloc(1,
-                                              sizeof(php_brotli_state_context));
-    ctx->encoder = NULL;
-    ctx->decoder = NULL;
-    return ctx;
-}
-
-static void php_brotli_state_destroy(php_brotli_state_context *ctx)
-{
-    if (ctx->encoder) {
-        BrotliEncoderDestroyInstance(ctx->encoder);
-        ctx->encoder = NULL;
-    }
-    if (ctx->decoder) {
-        BrotliDecoderDestroyInstance(ctx->decoder);
-        ctx->decoder = NULL;
-    }
-    efree(ctx);
-}
-
-static void php_brotli_state_rsrc_dtor(zend_resource *res)
-{
-    php_brotli_state_context *ctx = zend_fetch_resource(res, NULL, le_state);
-    php_brotli_state_destroy(ctx);
-}
-
 #define PHP_BROTLI_OUTPUT_HANDLER "ob_brotli_handler"
 
 static int php_brotli_output_encoding(void)
@@ -226,7 +186,7 @@ static int php_brotli_output_handler(void **handler_context,
                  != (PHP_OUTPUT_HANDLER_START
                      |PHP_OUTPUT_HANDLER_CLEAN|PHP_OUTPUT_HANDLER_FINAL))) {
             sapi_add_header_ex(ZEND_STRL("Vary: Accept-Encoding"),
-                               1, 0 TSRMLS_CC);
+                               1, 0);
         }
         return FAILURE;
     }
@@ -302,9 +262,9 @@ static int php_brotli_output_handler(void **handler_context,
             php_brotli_context_close(ctx);
 
             sapi_add_header_ex(ZEND_STRL("Content-Encoding: br"),
-                               1, 1 TSRMLS_CC);
+                               1, 1);
             sapi_add_header_ex(ZEND_STRL("Vary: Accept-Encoding"),
-                               1, 0 TSRMLS_CC);
+                               1, 0);
         }
     } else {
         php_brotli_context_close(ctx);
@@ -338,7 +298,7 @@ static php_brotli_context *php_brotli_output_handler_context_init(void)
     return ctx;
 }
 
-static void php_brotli_output_handler_context_dtor(void *opaq TSRMLS_DC)
+static void php_brotli_output_handler_context_dtor(void *opaq)
 {
     php_brotli_context *ctx = (php_brotli_context *)opaq;
 
@@ -361,11 +321,15 @@ php_brotli_output_handler_init(const char *handler_name,
 
     handler = php_output_handler_create_internal(handler_name, handler_name_len,
                                                  php_brotli_output_handler,
-                                                 chunk_size, flags TSRMLS_CC);
+                                                 chunk_size, flags);
     if (!handler) {
         return NULL;
     }
 
+    if (!BROTLI_G(output_compression)) {
+        BROTLI_G(output_compression) = 1;
+    }
+
     BROTLI_G(handler_registered) = 1;
 
     if (!BROTLI_G(ob_handler)) {
@@ -374,10 +338,7 @@ php_brotli_output_handler_init(const char *handler_name,
 
     php_output_handler_set_context(handler,
                                    BROTLI_G(ob_handler),
-                                   php_brotli_output_handler_context_dtor
-                                   TSRMLS_CC);
-
-    BROTLI_G(output_compression) = 1;
+                                   php_brotli_output_handler_context_dtor);
 
     return handler;
 }
@@ -386,7 +347,7 @@ static void php_brotli_cleanup_ob_handler_mess(void)
 {
     if (BROTLI_G(ob_handler)) {
         php_brotli_output_handler_context_dtor(
-            (void *) BROTLI_G(ob_handler) TSRMLS_CC
+            (void *) BROTLI_G(ob_handler)
         );
         BROTLI_G(ob_handler) = NULL;
     }
@@ -406,7 +367,7 @@ static void php_brotli_output_compression_start(void)
                     ZEND_STRL(PHP_BROTLI_OUTPUT_HANDLER),
                     PHP_OUTPUT_HANDLER_DEFAULT_SIZE,
                     PHP_OUTPUT_HANDLER_STDFLAGS))) {
-                php_output_handler_start(h TSRMLS_CC);
+                php_output_handler_start(h);
             }
             break;
     }
@@ -442,9 +403,9 @@ static PHP_INI_MH(OnUpdate_brotli_output_compression)
     }
 
     if (stage == PHP_INI_STAGE_RUNTIME) {
-        status = php_output_get_status(TSRMLS_C);
+        status = php_output_get_status();
         if (status & PHP_OUTPUT_SENT) {
-            php_error_docref("ref.outcontrol" TSRMLS_CC, E_WARNING,
+            php_error_docref("ref.outcontrol", E_WARNING,
                              "Cannot change brotli.output_compression"
                              " - headers already sent");
             return FAILURE;
@@ -454,9 +415,10 @@ static PHP_INI_MH(OnUpdate_brotli_output_compression)
     p = (zend_long *)(base+(size_t) mh_arg1);
     *p = int_value;
 
+    BROTLI_G(output_compression) = BROTLI_G(output_compression_default);
+
     if (stage == PHP_INI_STAGE_RUNTIME && int_value) {
-        if (!php_output_handler_started(ZEND_STRL(PHP_BROTLI_OUTPUT_HANDLER)
-                                        TSRMLS_CC)) {
+        if (!php_output_handler_started(ZEND_STRL(PHP_BROTLI_OUTPUT_HANDLER))) {
             php_brotli_output_compression_start();
         }
     }
@@ -464,15 +426,13 @@ static PHP_INI_MH(OnUpdate_brotli_output_compression)
     return SUCCESS;
 }
 
-static int php_brotli_output_conflict(const char *handler_name, size_t handler_name_len TSRMLS_DC)
+static int php_brotli_output_conflict(const char *handler_name, size_t handler_name_len)
 {
-    if (php_output_get_level(TSRMLS_C)) {
+    if (php_output_get_level()) {
         if (php_output_handler_conflict(handler_name, handler_name_len,
-                                        ZEND_STRL(PHP_BROTLI_OUTPUT_HANDLER)
-                                        TSRMLS_CC)
+                                        ZEND_STRL(PHP_BROTLI_OUTPUT_HANDLER))
             || php_output_handler_conflict(handler_name, handler_name_len,
-                                           ZEND_STRL("ob_gzhandler")
-                                           TSRMLS_CC)) {
+                                           ZEND_STRL("ob_gzhandler"))) {
             return FAILURE;
         }
     }
@@ -480,9 +440,10 @@ static int php_brotli_output_conflict(const char *handler_name, size_t handler_n
 }
 
 PHP_INI_BEGIN()
-  STD_PHP_INI_ENTRY("brotli.output_compression", "0",
-                    PHP_INI_ALL, OnUpdate_brotli_output_compression,
-                    output_compression, zend_brotli_globals, brotli_globals)
+  STD_PHP_INI_BOOLEAN("brotli.output_compression", "0",
+                      PHP_INI_ALL, OnUpdate_brotli_output_compression,
+                      output_compression_default,
+                      zend_brotli_globals, brotli_globals)
   STD_PHP_INI_ENTRY("brotli.output_compression_level", "-1",
                     PHP_INI_ALL, OnUpdateLong, output_compression_level,
                     zend_brotli_globals, brotli_globals)
@@ -513,7 +474,7 @@ typedef struct _php_brotli_stream_data {
 #define STREAM_NAME "compress.brotli"
 
 static int php_brotli_decompress_close(php_stream *stream,
-                                       int close_handle TSRMLS_DC)
+                                       int close_handle)
 {
     STREAM_DATA_FROM_STREAM();
 
@@ -545,13 +506,13 @@ static int php_brotli_decompress_close(php_stream *stream,
 #if PHP_VERSION_ID < 70400
 static size_t php_brotli_decompress_read(php_stream *stream,
                                          char *buf,
-                                         size_t count TSRMLS_DC)
+                                         size_t count)
 {
     size_t ret = 0;
 #else
 static ssize_t php_brotli_decompress_read(php_stream *stream,
                                          char *buf,
-                                         size_t count TSRMLS_DC)
+                                         size_t count)
 {
     ssize_t ret = 0;
 #endif
@@ -621,7 +582,7 @@ static ssize_t php_brotli_decompress_read(php_stream *stream,
 }
 
 static int php_brotli_compress_close(php_stream *stream,
-                                     int close_handle TSRMLS_DC)
+                                     int close_handle)
 {
     STREAM_DATA_FROM_STREAM();
 
@@ -649,8 +610,8 @@ static int php_brotli_compress_close(php_stream *stream,
                 php_stream_write(self->stream, output, out_size);
             }
         } else {
-            php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                             "brotli compress error\n");
+            php_error_docref(NULL, E_WARNING,
+                             "brotli compress error");
         }
     }
 
@@ -679,13 +640,13 @@ static int php_brotli_compress_close(php_stream *stream,
 #if PHP_VERSION_ID < 70400
 static size_t php_brotli_compress_write(php_stream *stream,
                                         const char *buf,
-                                        size_t count TSRMLS_DC)
+                                        size_t count)
 {
     size_t ret = 0;
 #else
 static ssize_t php_brotli_compress_write(php_stream *stream,
                                         const char *buf,
-                                        size_t count TSRMLS_DC)
+                                        size_t count)
 {
     ssize_t ret = 0;
 #endif
@@ -712,7 +673,7 @@ static ssize_t php_brotli_compress_write(php_stream *stream,
                 php_stream_write(self->stream, output, out_size);
             }
         } else {
-            php_error_docref(NULL TSRMLS_CC, E_WARNING, "brotli compress error\n");
+            php_error_docref(NULL, E_WARNING, "brotli compress error");
 #if PHP_VERSION_ID >= 70400
             return -1;
 #endif
@@ -756,13 +717,9 @@ php_stream_brotli_opener(
     const char *path,
     const char *mode,
     int options,
-#if PHP_MAJOR_VERSION < 7
-    char **opened_path,
-#else
     zend_string **opened_path,
-#endif
     php_stream_context *context
-    STREAMS_DC TSRMLS_DC)
+    STREAMS_DC)
 {
     php_brotli_stream_data *self;
     int level = BROTLI_DEFAULT_QUALITY;
@@ -775,7 +732,7 @@ php_stream_brotli_opener(
         }
     }
 
-    if (php_check_open_basedir(path TSRMLS_CC)) {
+    if (php_check_open_basedir(path)) {
         return NULL;
     }
 
@@ -784,30 +741,20 @@ php_stream_brotli_opener(
     } else if (!strcmp(mode, "r") || !strcmp(mode, "rb")) {
        compress = 0;
     } else {
-        php_error_docref(NULL TSRMLS_CC, E_ERROR, "brotli: invalid open mode");
+        php_error_docref(NULL, E_ERROR, "brotli: invalid open mode");
         return NULL;
     }
 
     if (context) {
-#if PHP_MAJOR_VERSION >= 7
         zval *tmpzval;
 
         if (NULL != (tmpzval = php_stream_context_get_option(
                          context, "brotli", "level"))) {
             level = zval_get_long(tmpzval);
         }
-#else
-        zval **tmpzval;
-
-        if (php_stream_context_get_option(
-                context, "brotli", "level", &tmpzval) == SUCCESS) {
-            convert_to_long_ex(tmpzval);
-            level = Z_LVAL_PP(tmpzval);
-        }
-#endif
     }
     if (level > BROTLI_MAX_QUALITY) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
+        php_error_docref(NULL, E_WARNING,
                          "brotli: compression level (%d) must be less than %d",
                          level, BROTLI_MAX_QUALITY);
         level = BROTLI_MAX_QUALITY;
@@ -825,7 +772,7 @@ php_stream_brotli_opener(
     if (compress) {
         self->dctx = NULL;
         if (php_brotli_encoder_create(&self->cctx, level, 0, 0) != SUCCESS) {
-            php_error_docref(NULL TSRMLS_CC, E_WARNING,
+            php_error_docref(NULL, E_WARNING,
                              "brotli: compression context failed");
             php_stream_close(self->stream);
             efree(self);
@@ -842,7 +789,7 @@ php_stream_brotli_opener(
     } else {
         self->cctx = NULL;
         if (php_brotli_decoder_create(&self->dctx) != SUCCESS) {
-            php_error_docref(NULL TSRMLS_CC, E_WARNING,
+            php_error_docref(NULL, E_WARNING,
                              "brotli: decompression context failed");
             php_stream_close(self->stream);
             efree(self);
@@ -882,6 +829,135 @@ php_stream_wrapper php_stream_brotli_wrapper = {
     0 /* is_url */
 };
 
+/* Brotli Compress/UnCompress Context */
+#define PHP_BROTLI_CONTEXT_OBJ_INIT_OF_CLASS(ce) \
+  php_brotli_context *ctx; \
+  object_init_ex(return_value, ce); \
+  ctx = php_brotli_context_from_obj(Z_OBJ_P(return_value)); \
+  ctx->state.encoder = NULL; \
+  ctx->state.decoder = NULL; \
+  ctx->available_in = 0; \
+  ctx->next_in = NULL; \
+  ctx->available_out = 0; \
+  ctx->next_out = NULL; \
+  ctx->output = NULL;
+
+static php_brotli_context *php_brotli_context_from_obj(zend_object *obj)
+{
+    return (php_brotli_context *)
+        ((char *)(obj) - XtOffsetOf(php_brotli_context, std));
+}
+
+static void php_brotli_context_free_obj(zend_object *object)
+{
+    php_brotli_context *intern = php_brotli_context_from_obj(object);
+    php_brotli_context_close(intern);
+    zend_object_std_dtor(&intern->std);
+}
+
+static zend_object *
+php_brotli_context_create_object(zend_class_entry *class_type,
+                                 zend_object_handlers *handlers)
+{
+    php_brotli_context *intern;
+#if PHP_VERSION_ID >= 80000
+    intern = zend_object_alloc(sizeof(php_brotli_context), class_type);
+#else
+    intern = ecalloc(1,
+                     sizeof(php_brotli_context)
+                     + zend_object_properties_size(class_type));
+#endif
+    zend_object_std_init(&intern->std, class_type);
+    object_properties_init(&intern->std, class_type);
+    intern->std.handlers = handlers;
+
+    return &intern->std;
+}
+
+/* Brotli Compress Context */
+zend_class_entry *php_brotli_compress_context_ce;
+static zend_object_handlers php_brotli_compress_context_object_handlers;
+
+static zend_object *
+php_brotli_compress_context_create_object(zend_class_entry *class_type)
+{
+    return php_brotli_context_create_object(
+        class_type,
+        &php_brotli_compress_context_object_handlers);
+}
+
+static zend_function *
+php_brotli_compress_context_get_constructor(zend_object *object)
+{
+    zend_throw_error(NULL,
+                     "Cannot directly construct Brotli\\Compress\\Context, "
+                     "use brotli_compress_init() instead");
+    return NULL;
+}
+
+static zend_class_entry *php_brotli_compress_context_register_class(void)
+{
+    zend_class_entry ce, *class_entry;
+
+    INIT_NS_CLASS_ENTRY(ce, "Brotli\\Compress", "Context", NULL);
+#if PHP_VERSION_ID >= 80000
+    class_entry = zend_register_internal_class_ex(&ce, NULL);
+#if PHP_VERSION_ID >= 80100
+    class_entry->ce_flags |= ZEND_ACC_FINAL|ZEND_ACC_NO_DYNAMIC_PROPERTIES|ZEND_ACC_NOT_SERIALIZABLE;
+#else
+    class_entry->ce_flags |= ZEND_ACC_FINAL|ZEND_ACC_NO_DYNAMIC_PROPERTIES;
+#endif
+#else
+    ce.create_object = php_brotli_compress_context_create_object;
+    class_entry = zend_register_internal_class(&ce);
+    class_entry->ce_flags |= ZEND_ACC_FINAL;
+#endif
+
+    return class_entry;
+}
+
+/* Brotli UnCompress Context */
+zend_class_entry *php_brotli_uncompress_context_ce;
+static zend_object_handlers php_brotli_uncompress_context_object_handlers;
+
+static zend_object *
+php_brotli_uncompress_context_create_object(zend_class_entry *class_type)
+{
+    return php_brotli_context_create_object(
+        class_type,
+        &php_brotli_uncompress_context_object_handlers);
+}
+
+static zend_function *
+php_brotli_uncompress_context_get_constructor(zend_object *object)
+{
+  zend_throw_error(NULL,
+                   "Cannot directly construct Brotli\\UnCompress\\Context, "
+                   "use brotli_uncompress_init() instead");
+  return NULL;
+}
+
+static zend_class_entry *php_brotli_uncompress_context_register_class(void)
+{
+    zend_class_entry ce, *class_entry;
+
+    INIT_NS_CLASS_ENTRY(ce, "Brotli\\UnCompress", "Context", NULL);
+#if PHP_VERSION_ID >= 80000
+    class_entry = zend_register_internal_class_ex(&ce, NULL);
+#if PHP_VERSION_ID >= 80100
+    class_entry->ce_flags |= ZEND_ACC_FINAL|ZEND_ACC_NO_DYNAMIC_PROPERTIES|ZEND_ACC_NOT_SERIALIZABLE;
+#else
+    class_entry->ce_flags |= ZEND_ACC_FINAL|ZEND_ACC_NO_DYNAMIC_PROPERTIES;
+#endif
+#else
+    ce.create_object = php_brotli_uncompress_context_create_object;
+    class_entry = zend_register_internal_class(&ce);
+    class_entry->ce_flags |= ZEND_ACC_FINAL;
+#endif
+
+    return class_entry;
+}
+
 ZEND_MINIT_FUNCTION(brotli)
 {
     ZEND_INIT_MODULE_GLOBALS(brotli, php_brotli_init_globals, NULL);
@@ -908,19 +984,71 @@ ZEND_MINIT_FUNCTION(brotli)
     REGISTER_LONG_CONSTANT("BROTLI_FINISH", BROTLI_OPERATION_FINISH,
                            CONST_CS | CONST_PERSISTENT);
 
-    le_state = zend_register_list_destructors_ex(php_brotli_state_rsrc_dtor,
-                                                 NULL, "brotli.state",
-                                                 module_number);
-
     php_output_handler_alias_register(ZEND_STRL(PHP_BROTLI_OUTPUT_HANDLER),
-                                      php_brotli_output_handler_init TSRMLS_CC);
+                                      php_brotli_output_handler_init);
     php_output_handler_conflict_register(ZEND_STRL(PHP_BROTLI_OUTPUT_HANDLER),
-                                         php_brotli_output_conflict TSRMLS_CC);
+                                         php_brotli_output_conflict);
+
+    php_brotli_compress_context_ce
+        = php_brotli_compress_context_register_class();
+#if PHP_VERSION_ID >= 80000
+    php_brotli_compress_context_ce->create_object
+        = php_brotli_compress_context_create_object;
+#if PHP_VERSION_ID >= 80300
+    php_brotli_compress_context_ce->default_object_handlers
+        = &php_brotli_compress_context_object_handlers;
+#endif
+#if PHP_VERSION_ID < 80100
+    php_brotli_compress_context_ce->serialize = zend_class_serialize_deny;
+    php_brotli_compress_context_ce->unserialize = zend_class_unserialize_deny;
+#endif
+#endif
+    memcpy(&php_brotli_compress_context_object_handlers,
+           &std_object_handlers, sizeof(zend_object_handlers));
+    php_brotli_compress_context_object_handlers.offset
+        = XtOffsetOf(php_brotli_context, std);
+    php_brotli_compress_context_object_handlers.free_obj
+        = php_brotli_context_free_obj;
+    php_brotli_compress_context_object_handlers.get_constructor
+        = php_brotli_compress_context_get_constructor;
+    php_brotli_compress_context_object_handlers.clone_obj = NULL;
+#if PHP_VERSION_ID >= 80000
+    php_brotli_compress_context_object_handlers.compare
+        = zend_objects_not_comparable;
+#endif
+
+    php_brotli_uncompress_context_ce
+        = php_brotli_uncompress_context_register_class();
+#if PHP_VERSION_ID >= 80000
+    php_brotli_uncompress_context_ce->create_object
+        = php_brotli_uncompress_context_create_object;
+#if PHP_VERSION_ID >= 80300
+    php_brotli_uncompress_context_ce->default_object_handlers
+        = &php_brotli_uncompress_context_object_handlers;
+#endif
+#if PHP_VERSION_ID < 80100
+    php_brotli_uncompress_context_ce->serialize = zend_class_serialize_deny;
+    php_brotli_uncompress_context_ce->unserialize = zend_class_unserialize_deny;
+#endif
+#endif
+    memcpy(&php_brotli_uncompress_context_object_handlers,
+           &std_object_handlers, sizeof(zend_object_handlers));
+    php_brotli_uncompress_context_object_handlers.offset
+        = XtOffsetOf(php_brotli_context, std);
+    php_brotli_uncompress_context_object_handlers.free_obj
+        = php_brotli_context_free_obj;
+    php_brotli_uncompress_context_object_handlers.get_constructor
+        = php_brotli_uncompress_context_get_constructor;
+    php_brotli_uncompress_context_object_handlers.clone_obj = NULL;
+#if PHP_VERSION_ID >= 80000
+    php_brotli_uncompress_context_object_handlers.compare
+        = zend_objects_not_comparable;
+#endif
 
     REGISTER_INI_ENTRIES();
 
     php_register_url_stream_wrapper(STREAM_NAME,
-                                    &php_stream_brotli_wrapper TSRMLS_CC);
+                                    &php_stream_brotli_wrapper);
 
 #if defined(HAVE_APCU_SUPPORT)
     apc_register_serializer("brotli",
@@ -942,6 +1070,7 @@ ZEND_RINIT_FUNCTION(brotli)
 {
     BROTLI_G(compression_coding) = 0;
     if (!BROTLI_G(handler_registered)) {
+        BROTLI_G(output_compression) = BROTLI_G(output_compression_default);
         php_brotli_output_compression_start();
     }
     return SUCCESS;
@@ -949,9 +1078,7 @@ ZEND_RINIT_FUNCTION(brotli)
 
 ZEND_RSHUTDOWN_FUNCTION(brotli)
 {
-    if (BROTLI_G(handler_registered)) {
-        php_brotli_cleanup_ob_handler_mess();
-    }
+    php_brotli_cleanup_ob_handler_mess();
     BROTLI_G(handler_registered) = 0;
     return SUCCESS;
 }
@@ -988,7 +1115,7 @@ zend_module_entry brotli_module_entry = {
     STANDARD_MODULE_HEADER_EX,
     NULL,
     brotli_module_deps,
-#elif ZEND_MODULE_API_NO >= 20010901
+#else
     STANDARD_MODULE_HEADER,
 #endif
     "brotli",
@@ -998,14 +1125,12 @@ zend_module_entry brotli_module_entry = {
     ZEND_RINIT(brotli),
     ZEND_RSHUTDOWN(brotli),
     ZEND_MINFO(brotli),
-#if ZEND_MODULE_API_NO >= 20010901
     BROTLI_EXT_VERSION,
-#endif
     STANDARD_MODULE_PROPERTIES
 };
 
 #ifdef COMPILE_DL_BROTLI
-#if PHP_MAJOR_VERSION >= 7 && defined(ZTS)
+#if defined(ZTS)
 ZEND_TSRMLS_CACHE_DEFINE()
 #endif
 ZEND_GET_MODULE(brotli)
@@ -1014,19 +1139,17 @@ ZEND_GET_MODULE(brotli)
 static ZEND_FUNCTION(brotli_compress)
 {
     char *in;
-#if ZEND_MODULE_API_NO >= 20141001
     size_t in_size;
-#else
-    int in_size;
-#endif
-    long quality = BROTLI_DEFAULT_QUALITY;
-    long mode =  BROTLI_MODE_GENERIC;
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
-                              "s|ll", &in, &in_size,
-                              &quality, &mode) == FAILURE) {
-        RETURN_FALSE;
-    }
+    zend_long quality = BROTLI_DEFAULT_QUALITY;
+    zend_long mode =  BROTLI_MODE_GENERIC;
+
+    ZEND_PARSE_PARAMETERS_START(1, 3)
+        Z_PARAM_STRING(in, in_size)
+        Z_PARAM_OPTIONAL
+        Z_PARAM_LONG(quality)
+        Z_PARAM_LONG(mode)
+    ZEND_PARSE_PARAMETERS_END();
 
     size_t out_size = BrotliEncoderMaxCompressedSize(in_size);
     char *out = (char *)emalloc(out_size);
@@ -1046,62 +1169,72 @@ static ZEND_FUNCTION(brotli_compress)
     if (!BrotliEncoderCompress((int)quality, lgwin, (BrotliEncoderMode)mode,
                                in_size, (const uint8_t*)in,
                                &out_size, (uint8_t*)out)) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                         "Brotli compress failed\n");
+        php_error_docref(NULL, E_WARNING,
+                         "Brotli compress failed");
         efree(out);
         RETURN_FALSE;
     }
 
-#if ZEND_MODULE_API_NO >= 20141001
     RETVAL_STRINGL(out, out_size);
-#else
-    RETVAL_STRINGL(out, out_size, 1);
-#endif
     efree(out);
 }
 
 static ZEND_FUNCTION(brotli_compress_init)
 {
-    long quality = BROTLI_DEFAULT_QUALITY;
-    long mode =  BROTLI_MODE_GENERIC;
-    php_brotli_state_context *ctx;
+    zend_long quality = BROTLI_DEFAULT_QUALITY;
+    zend_long mode =  BROTLI_MODE_GENERIC;
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
-                              "|ll", &quality, &mode) == FAILURE) {
-        RETURN_FALSE;
-    }
+    ZEND_PARSE_PARAMETERS_START(0, 2)
+        Z_PARAM_OPTIONAL
+        Z_PARAM_LONG(quality)
+        Z_PARAM_LONG(mode)
+    ZEND_PARSE_PARAMETERS_END();
 
-    ctx = php_brotli_state_init();
+    PHP_BROTLI_CONTEXT_OBJ_INIT_OF_CLASS(php_brotli_compress_context_ce);
 
-    if (php_brotli_encoder_create(&ctx->encoder,
+    if (php_brotli_encoder_create(&ctx->state.encoder,
                                   quality, 0, mode) != SUCCESS) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                         "Brotli incremental compress init failed\n");
+        zval_ptr_dtor(return_value);
+        php_error_docref(NULL, E_WARNING,
+                         "Brotli incremental compress init failed");
         RETURN_FALSE;
     }
-
-    RETURN_RES(zend_register_resource(ctx, le_state));
 }
 
 static ZEND_FUNCTION(brotli_compress_add)
 {
     zval *res;
-    php_brotli_state_context *ctx;
+    php_brotli_context *ctx;
     size_t buffer_size, buffer_used;
     zend_long mode = BROTLI_OPERATION_FLUSH;
+#if PHP_VERSION_ID >= 80000
+    zend_object *obj;
+#else
+    zval *obj;
+#endif
     char *in_buf;
     size_t in_size;
     smart_string out = {0};
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS(), "rs|l",
-                              &res, &in_buf, &in_size, &mode) != SUCCESS) {
-        RETURN_FALSE;
-    }
+    ZEND_PARSE_PARAMETERS_START(2, 3)
+#if PHP_VERSION_ID >= 80000
+        Z_PARAM_OBJ_OF_CLASS(obj, php_brotli_compress_context_ce)
+#else
+        Z_PARAM_OBJECT_OF_CLASS(obj, php_brotli_compress_context_ce)
+#endif
+        Z_PARAM_STRING(in_buf, in_size)
+        Z_PARAM_OPTIONAL
+        Z_PARAM_LONG(mode)
+    ZEND_PARSE_PARAMETERS_END();
 
-    ctx = zend_fetch_resource(Z_RES_P(res), NULL, le_state);
-    if (ctx == NULL || ctx->encoder == NULL) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                         "Brotli incremental compress resource failed\n");
+#if PHP_VERSION_ID >= 80000
+    ctx = php_brotli_context_from_obj(obj);
+#else
+    ctx = php_brotli_context_from_obj(Z_OBJ_P(obj));
+#endif
+    if (ctx == NULL || ctx->state.encoder == NULL) {
+        php_error_docref(NULL, E_WARNING,
+                         "Brotli incremental compress resource failed");
         RETURN_FALSE;
     }
 
@@ -1109,52 +1242,53 @@ static ZEND_FUNCTION(brotli_compress_add)
     buffer_size = (buffer_size < 64) ? 64 : buffer_size;
     uint8_t *buffer = (uint8_t *)emalloc(buffer_size);
 
-    const uint8_t *next_in = in_buf;
-    size_t available_in = in_size;
+    ctx->next_in = in_buf;
+    ctx->available_in = in_size;
 
-    while (available_in || BrotliEncoderHasMoreOutput(ctx->encoder)) {
-        size_t available_out = buffer_size;
-        uint8_t *next_out = buffer;
-        if (BrotliEncoderCompressStream(ctx->encoder,
+    while (ctx->available_in
+           || BrotliEncoderHasMoreOutput(ctx->state.encoder)) {
+        ctx->available_out = buffer_size;
+        ctx->next_out = buffer;
+        if (BrotliEncoderCompressStream(ctx->state.encoder,
                                         mode,
-                                        &available_in,
-                                        &next_in,
-                                        &available_out,
-                                        &next_out,
+                                        &ctx->available_in,
+                                        &ctx->next_in,
+                                        &ctx->available_out,
+                                        &ctx->next_out,
                                         0)) {
-            buffer_used = (size_t)(next_out - buffer);
+            buffer_used = (size_t)(ctx->next_out - buffer);
             if (buffer_used) {
                 smart_string_appendl(&out, buffer, buffer_used);
             }
         } else {
             efree(buffer);
             smart_string_free(&out);
-            php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                             "Brotli incremental compress failed\n");
+            php_error_docref(NULL, E_WARNING,
+                             "Brotli incremental compress failed");
             RETURN_FALSE;
         }
     }
 
     if (mode == BROTLI_OPERATION_FINISH) {
-        while (!BrotliEncoderIsFinished(ctx->encoder)) {
-            size_t available_out = buffer_size;
-            uint8_t *next_out = buffer;
-            if (BrotliEncoderCompressStream(ctx->encoder,
+        while (!BrotliEncoderIsFinished(ctx->state.encoder)) {
+            ctx->available_out = buffer_size;
+            ctx->next_out = buffer;
+            if (BrotliEncoderCompressStream(ctx->state.encoder,
                                             BROTLI_OPERATION_FINISH,
-                                            &available_in,
-                                            &next_in,
-                                            &available_out,
-                                            &next_out,
+                                            &ctx->available_in,
+                                            &ctx->next_in,
+                                            &ctx->available_out,
+                                            &ctx->next_out,
                                             0)) {
-                buffer_used = (size_t)(next_out - buffer);
+                buffer_used = (size_t)(ctx->next_out - buffer);
                 if (buffer_used) {
                     smart_string_appendl(&out, buffer, buffer_used);
                 }
             } else {
                 efree(buffer);
                 smart_string_free(&out);
-                php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                                 "Brotli incremental compress failed\n");
+                php_error_docref(NULL, E_WARNING,
+                                 "Brotli incremental compress failed");
                 RETURN_FALSE;
             }
         }
@@ -1168,20 +1302,16 @@ static ZEND_FUNCTION(brotli_compress_add)
 
 static ZEND_FUNCTION(brotli_uncompress)
 {
-    long max_size = 0;
+    zend_long max_size = 0;
     char *in;
-#if ZEND_MODULE_API_NO >= 20141001
     size_t in_size;
     smart_string out = {0};
-#else
-    int in_size;
-    smart_str out = {0};
-#endif
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l",
-                              &in, &in_size, &max_size) == FAILURE) {
-        RETURN_FALSE;
-    }
+    ZEND_PARSE_PARAMETERS_START(1, 2)
+        Z_PARAM_STRING(in, in_size)
+        Z_PARAM_OPTIONAL
+        Z_PARAM_LONG(max_size)
+    ZEND_PARSE_PARAMETERS_END();
 
     if (max_size && max_size < in_size) {
         in_size = max_size;
@@ -1189,8 +1319,8 @@ static ZEND_FUNCTION(brotli_uncompress)
 
     BrotliDecoderState *state = BrotliDecoderCreateInstance(NULL, NULL, NULL);
     if (!state) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                         "Invalid Brotli state\n");
+        php_error_docref(NULL, E_WARNING,
+                         "Invalid Brotli state");
         RETURN_FALSE;
     }
 
@@ -1208,11 +1338,7 @@ static ZEND_FUNCTION(brotli_uncompress)
                                                0);
         size_t used_out = buffer_size - available_out;
         if (used_out != 0) {
-#if ZEND_MODULE_API_NO >= 20141001
             smart_string_appendl(&out, buffer, used_out);
-#else
-            smart_str_appendl(&out, buffer, used_out);
-#endif
         }
     }
 
@@ -1220,68 +1346,65 @@ static ZEND_FUNCTION(brotli_uncompress)
     efree(buffer);
 
     if (result != BROTLI_DECODER_RESULT_SUCCESS) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                         "Brotli decompress failed\n");
-#if ZEND_MODULE_API_NO >= 20141001
+        php_error_docref(NULL, E_WARNING,
+                         "Brotli decompress failed");
         smart_string_free(&out);
-#else
-        smart_str_free(&out);
-#endif
         RETURN_FALSE;
     }
 
-#if ZEND_MODULE_API_NO >= 20141001
     RETVAL_STRINGL(out.c, out.len);
-#else
-    RETVAL_STRINGL(out.c, out.len, 1);
-#endif
-
-#if ZEND_MODULE_API_NO >= 20141001
     smart_string_free(&out);
-#else
-    smart_str_free(&out);
-#endif
 }
 
-#if PHP_VERSION_ID >= 70000
 static ZEND_FUNCTION(brotli_uncompress_init)
 {
-    php_brotli_state_context *ctx;
-
     if (zend_parse_parameters_none() == FAILURE) {
         RETURN_FALSE;
     }
 
-    ctx = php_brotli_state_init();
+    PHP_BROTLI_CONTEXT_OBJ_INIT_OF_CLASS(php_brotli_uncompress_context_ce);
 
-    if (php_brotli_decoder_create(&ctx->decoder) != SUCCESS) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                         "Brotli incremental uncompress init failed\n");
+    if (php_brotli_decoder_create(&ctx->state.decoder) != SUCCESS) {
+        php_error_docref(NULL, E_WARNING,
+                         "Brotli incremental uncompress init failed");
         RETURN_FALSE;
     }
-
-    RETURN_RES(zend_register_resource(ctx, le_state));
 }
 
 static ZEND_FUNCTION(brotli_uncompress_add)
 {
     zval *res;
-    php_brotli_state_context *ctx;
+    php_brotli_context *ctx;
     size_t buffer_size;
     zend_long mode = BROTLI_OPERATION_FLUSH;
+#if PHP_VERSION_ID >= 80000
+    zend_object *obj;
+#else
+    zval *obj;
+#endif
     char *in_buf;
     size_t in_size;
     smart_string out = {0};
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS(), "rs|l",
-                              &res, &in_buf, &in_size, &mode) != SUCCESS) {
-        RETURN_FALSE;
-    }
+    ZEND_PARSE_PARAMETERS_START(2, 3)
+#if PHP_VERSION_ID >= 80000
+        Z_PARAM_OBJ_OF_CLASS(obj, php_brotli_uncompress_context_ce)
+#else
+        Z_PARAM_OBJECT_OF_CLASS(obj, php_brotli_uncompress_context_ce)
+#endif
+        Z_PARAM_STRING(in_buf, in_size)
+        Z_PARAM_OPTIONAL
+        Z_PARAM_LONG(mode)
+    ZEND_PARSE_PARAMETERS_END();
 
-    ctx = zend_fetch_resource(Z_RES_P(res), NULL, le_state);
-    if (ctx == NULL || ctx->decoder == NULL) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING,
-                         "Brotli incremental uncompress resource failed\n");
+#if PHP_VERSION_ID >= 80000
+    ctx = php_brotli_context_from_obj(obj);
+#else
+    ctx = php_brotli_context_from_obj(Z_OBJ_P(obj));
+#endif
+    if (ctx == NULL || ctx->state.decoder == NULL) {
+        php_error_docref(NULL, E_WARNING,
+                         "Brotli incremental uncompress resource failed");
         RETURN_FALSE;
     }
 
@@ -1292,18 +1415,20 @@ static ZEND_FUNCTION(brotli_uncompress_add)
     buffer_size = PHP_BROTLI_BUFFER_SIZE;
     uint8_t *buffer = (uint8_t *)emalloc(buffer_size);
 
-    const uint8_t *next_in = (const uint8_t *)in_buf;
-    size_t available_in = in_size;
+    ctx->next_in = (const uint8_t *)in_buf;
+    ctx->available_in = in_size;
 
     BrotliDecoderResult result = BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT;
     while (result == BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT) {
-        size_t available_out = buffer_size;
-        uint8_t *next_out = buffer;
-        result = BrotliDecoderDecompressStream(ctx->decoder,
-                                               &available_in, &next_in,
-                                               &available_out, &next_out,
+        ctx->available_out = buffer_size;
+        ctx->next_out = buffer;
+        result = BrotliDecoderDecompressStream(ctx->state.decoder,
+                                               &ctx->available_in,
+                                               &ctx->next_in,
+                                               &ctx->available_out,
+                                               &ctx->next_out,
                                                0);
-        size_t buffer_used = buffer_size - available_out;
+        size_t buffer_used = buffer_size - ctx->available_out;
         if (buffer_used) {
             smart_string_appendl(&out, buffer, buffer_used);
         }
@@ -1314,7 +1439,6 @@ static ZEND_FUNCTION(brotli_uncompress_add)
     efree(buffer);
     smart_string_free(&out);
 }
-#endif
 
 #if defined(HAVE_APCU_SUPPORT)
 static int APC_SERIALIZER_NAME(brotli)(APC_SERIALIZER_ARGS)
@@ -1403,7 +1527,8 @@ static int APC_UNSERIALIZER_NAME(brotli)(APC_UNSERIALIZER_ARGS)
 
     if (!result) {
         php_error_docref(NULL, E_NOTICE,
-                         "Error at offset %ld of %ld bytes",
+                         "Error at offset " ZEND_LONG_FMT
+                         " of " ZEND_LONG_FMT " bytes",
                          (zend_long) ((char*) tmp - ZSTR_VAL(out.s)),
                          (zend_long) ZSTR_LEN(out.s));
         ZVAL_NULL(value);
