@@ -6,7 +6,7 @@ extern "C" {
 #endif
 
 /* Don't forget to check BROTLI_LIB_VERSION in config.m4/w32 */
-#define BROTLI_EXT_VERSION "0.15.0"
+#define BROTLI_EXT_VERSION "0.16.0"
 #define BROTLI_NS "Brotli"
 
 /* brotli */
@@ -40,10 +40,12 @@ typedef struct _php_brotli_context {
     size_t available_out;
     uint8_t *next_out;
     uint8_t *output;
+    zend_object std;
 } php_brotli_context;
 
 ZEND_BEGIN_MODULE_GLOBALS(brotli)
   zend_long output_compression;
+  zend_long output_compression_default;
   zend_long output_compression_level;
   zend_bool handler_registered;
   int compression_coding;
